@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 1999-2000 Free Software Foundation, Inc.
+ * Copyright (C) 1999-2001 Free Software Foundation, Inc.
  * This file is part of the GNU LIBICONV Library.
  *
  * The GNU LIBICONV Library is free software; you can redistribute it
@@ -1016,7 +1016,7 @@ uhc_2_wctomb (conv_t conv, unsigned char *r, ucs4_t wc, int n)
         return 2;
       }
     }
-    return RET_ILSEQ;
+    return RET_ILUNI;
   }
   return RET_TOOSMALL;
 }
